/*
	PROforma, PROGS FOnt & Raster MAnager

	PFCharAvailable gstate, unicode
	PFWidth gstate, string, *width
	PFWidthKern gstate, string, *width
	PFFontBbox gstate, pt *bbox
	PFPFontBbox gstate, code, *bbox

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 September 1993
	12.10.1995 PFFontBbox no longer needs CurrFontMatrix but calculates it on
			   the spot. This matrix was not used anywhere else.
			   Changes to do with better cache.
	16.10.1995 some changes because there is always a currfont
	02.12.1996 PFPFontBbox added
	05.11.2000 take fontsizemultiplier into account (affects PFWidth,
			   PFWidthKern, PFFontBbox)
	22.12.2002 KernAdjustWidth now also takes fontsizemultiplier into account
			   PCharWidth had efficiency mistake
*/

#include <array.h>

#include "gstate.h"
#include "system.h"
#include "path.h"
#include "hint.h"
#include "cache.h"

/* the builting font !! */
extern font *ResFonts;				/* list of resident fonts */

/* routine to get the width of a given character from given font - no gstate */
/* uses the limitation that either the hsbw or sbw command has to be the */
/* first command in a character */
/* if this is not true, a character width of 0 is returned */
/* the returned character width is the integer x width for a 1000pt font */
int glyphwidth(font *f, char *base)
{
	unsigned char command;
	int value;
	int stack[font_stack_size];
	int *currstack=stack;
	coor c,c1,c2;
	CTM *m=(CTM *)f->info->FontMatrix;

#define width(X,Y) pt2rshort( ((X)*m->xs-(Y)*m->xr)*1000 )
#define fwidth(X)  pt2rshort( ((X)*m->xs		  )*1000 )
#define getcommand command=(unsigned char)*base++;

	/* process character */
	while (TRUE)
	{
		getcommand;
		if (command>=minval)
		{
			if (command<=maxval)
			{
				*currstack++=command-139;
			} else if (command<=maxposword)
			{
				*currstack++=(command-minposword)*256 +
							 ((unsigned char)*base++) + 108;
			} else if (command<=maxnegword)
			{
				*currstack++=-(command-minnegword)*256 -
							  ((unsigned char)*base++) - 108;
			}
			else
			{
				MEMCopy(4,base,&value); /* !!!!!!!!!!!!! tricky !!!!!!!!! */
				base+=4;				/* sequence needs big endian long */
				*currstack++=value;
			}
		} else switch(command)
		{
		case fc_escape:
			getcommand;
			switch(command)
			{
			case fec_sbw:
				return width(stack[2],stack[3]);
			default:
				return 0;
			}
		case fc_hsbw:
			return fwidth(stack[1]);
		}
	}
}


/* compare routine for binary search */
static int compare(const void *a, const void *b)
{
	return ((long)(((ref *)a)->code)&0xffff) -
		   ((long)(((ref *)b)->code)&0xffff);
}

/* routine to see if given character exists */
Error PFCharAvailable(Gstate gstate, int unicode)
{
	/* search the given character code with a binary search */
	Error err;
	ref *rf, r;
	r.code=unicode;

	err=ARRAYSearch(&r, gstate->currfont->glyphs->list,
						gstate->currfont->glyphs->count,
						sizeof(ref), compare, &rf);
	return err?ERR_ITNF:ERR_OK;
}

/* get width of character with given unicode */
int CharWidth(Gstate gstate, int code)
{
	font_width *widths=gstate->currfont->widths;
	Error err;
	widthkey r, *rf;

	if (code<=PF_WIDTHARRAY_LAST && code>=PF_WIDTHARRAY_FIRST)
		return widths->array[code-PF_WIDTHARRAY_FIRST];

	r.unicode=code;
	err=ARRAYSearch(&r, widths->list, widths->count,
						sizeof(widthkey), compare, &rf);
	if (!err)
		return rf->width;

	/* haven't found character width yet -> search in builtin font */
	if (!ResFonts) return 0;
	widths=ResFonts->widths;
	if (code<=PF_WIDTHARRAY_LAST && code>=PF_WIDTHARRAY_FIRST)
		return widths->array[code-PF_WIDTHARRAY_FIRST];

	r.unicode=code;
	err=ARRAYSearch(&r, widths->list, widths->count,
						sizeof(widthkey), compare, &rf);
	if (!err)
		return rf->width;
	return 0;
}

/* get pixel width of character with given unicode */
int PCharWidth(Gstate gstate, int code)
{
	short width;
	font_width *widths=&gstate->font->pwidth;
	Error err;

	if (code<=PF_WIDTHARRAY_LAST && code>=PF_WIDTHARRAY_FIRST)
	{
		width=widths->array[code-PF_WIDTHARRAY_FIRST];
		if (width!=-1) return width;
	}

	gstate->noDisplayChars=TRUE;
	if (!gstate->font->mwidth) PFShowChar(gstate, 'm');
	PFShowChar(gstate, code);
	gstate->noDisplayChars=FALSE;
	PFPathClear(gstate);
	return pt2short(gstate->chwx);
}

/* compare routine for binary search */
int Kcompare(const void *a, const void *b)
{
	long d;
	d=((long)(((kernkey *)a)->leftcode)&0xffff)-
	  ((long)(((kernkey *)b)->leftcode)&0xffff);
	if (d)
		return d;
	else
		return ((long)(((kernkey *)a)->rightcode)&0xffff)-
			   ((long)(((kernkey *)b)->rightcode)&0xffff);
}

/* routine to get the amount of adjusting necessary for kerning */
int KernAdjust(Gstate gstate, short leftchar, short rightchar)
{
	/* search the given character code with a binary search */
	Error err;
	kernkey *rf, r;
	font_kern *kern;

	if (!(kern=gstate->currfont->kern)) return 0;

	r.leftcode=leftchar;
	r.rightcode=rightchar;

	err=ARRAYSearch(&r, kern->list, kern->count,
						sizeof(kernkey), Kcompare, &rf);
	if (err) return 0;
	return rf->adjust;
}

/* routine to get the adjust charwidth because of kerning */
void KernAdjustWidth(Gstate gstate, short leftchar, short rightchar)
{
	/* search the given character code with a binary search */
	Error err;
	kernkey *rf, r;
	font_kern *kern;

	if (!(kern=gstate->currfont->kern)) return;

	r.leftcode=leftchar;
	r.rightcode=rightchar;

	err=ARRAYSearch(&r, kern->list, kern->count,
						sizeof(kernkey), Kcompare, &rf);
	if (err) return;
	gstate->CharWidth.x+=rf->adjust*(gstate->fontsize/1000/gstate->fontsizemultiplier);
}

Error PFWidth(Gstate gstate, const short *string, pt *xwidth)
{
	int width=0;

	*xwidth=0;
	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		width+=CharWidth(gstate, *string);
		string++;
	}
	*xwidth=width*(gstate->fontsize/1000/gstate->fontsizemultiplier);
	return ERR_OK;
}

Error PFWidthKern(Gstate gstate, const short *string, pt *xwidth)
{
	int width=0;

	*xwidth=0;
	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		width+=CharWidth(gstate, *string);
		width+=KernAdjust(gstate, *string, *(string+1));
		string++;
	}
	*xwidth=width*(gstate->fontsize/1000/gstate->fontsizemultiplier);
	return ERR_OK;
}

/* enquire FontBbox, return size of bbox and offset of start of char in bbox */
Error PFFontBbox(Gstate gstate, Bbox *bbox)
{
	pt xoff, yoff, xsiz, ysiz;
	pt fontsize=gstate->fontsize/gstate->fontsizemultiplier;
	CTM fm; 						/* transformation, char to normal */
	CTM *om=(CTM *)gstate->currfont->info->FontMatrix;
	short *fb=gstate->currfont->info->FontBbox;

	/* calculate needed transformation factors */
	fm.xs=fixmul( om->xs, fontsize);
	fm.ys=fixmul( om->ys, fontsize);
	fm.xr=fixmul(-om->xr, fontsize);
	fm.yr=fixmul(-om->yr, fontsize);

	xoff= -fb[0]*fm.xs - fb[1]*fm.xr;
	yoff= -fb[1]*fm.ys - fb[0]*fm.yr;
	xsiz=  fb[2]*fm.xs + fb[3]*fm.xr;
	ysiz=  fb[3]*fm.ys + fb[2]*fm.yr;
	bbox->xsiz=xsiz+xoff;
	bbox->ysiz=ysiz+yoff;
	bbox->xorg=xoff;
	bbox->yorg=ysiz;
	return ERR_OK;
}

Error PFPFontBbox(Gstate gstate, short code, PBbox *bbox)
{
	int xmin=0x7fff,ymin=0x7fff;
	int xmax=-0x7fff,ymax=-0x7fff;
	int i;
	short list[6]={'W','f','g','m','_'};
	list[5]=code;

	if (!gstate->font) CacheFontFind(gstate);

	gstate->noDisplayChars=TRUE;
	for (i=0 ; i<6 && list[i] ; i++)
	{
		PFShowChar(gstate,list[i]);
		newmin(xmin,gstate->boxxmin);
		newmin(ymin,gstate->boxymin);
		newmax(xmax,gstate->boxxmax);
		newmax(ymax,gstate->boxymax);
	}
	gstate->noDisplayChars=FALSE;
	bbox->xsiz=xmax-xmin;
	bbox->ysiz=ymax-ymin;
	bbox->xorg=-xmin;
	bbox->yorg=-ymin;
	return ERR_OK;
}
