/*
	header file for PROforma
	definition of a Gstate object

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 10 July 1993

	19.03.1993 adjusted for (external) drivers
	Sep/Oct 1995, various modifications for v1.10
	09.11.1996 shuffled some parts of the Gstate to allow better information
			   hiding towards the PROforma external modules
	04.11.2000 fontsizemultiplier added for anti-aliased fonts
*/

#include <sms.h>
#include <win.h>
#include <mem.h>
#include <mutex.h>

#ifndef _PROforma_H
#include "PROforma.h"
#endif _PROforma_H

#ifndef catch
#define catch(x) do { if (err=(x)) return err; } while (0)
#endif

#define PF_LINE_UP			1	/* line fromend to start */
#define PF_LINE_DOWN		-1	/* line from start to end */

#define PF_PATH_HAIRLINE	5	/* path is too thin to be stroked ... */
#define PF_PATH_THICKLINE	6	/* path is too thick for hairline stroking ... */
								/* ... paths can automatically switch ... */
								/* ... between stroked, thickline and hairline */

/* routines and defines to draw filled paths */
#define PF_WINDING_RULE 	-1	/* use the winding rule */
#define PF_EVEN_ODD_RULE	1	/* use the even/odd rule */

#define BLOCK_NUMBER	300 	/* number of element in one segment block */

typedef struct {
		pt x,y; 				/* x and y coordinates */
} coor;

/* segment lists don't contain horizontal lines, and fall intirely inside */
/* the PageBbox. The current path always has the prev element pointing */
/* to the sentinel, for easy removal. */
typedef struct segment_ {
		struct segment_ *next;	/* pointer to next */
		struct segment_ *prev;	/* pointer to previous */
		struct segment_ *clip;	/* pointer to next element of clipping path */
		short starty, endy; 	/* start and end y coodinates (rounded) */
		short nexty;			/* next event point for this segment */
		pt startx, incr;		/* start x value and increment */
		char direction; 		/* direction of line */
		char isclip;			/* part of clipping path ?? */
} segment;

typedef struct segbl_ {
		struct segbl_ *next;	/* pointer to next segblock */
		segment list[BLOCK_NUMBER];
} segblock;

typedef struct _Bitmap *Bitmap;

typedef struct {
		short xsiz, ysiz;		/* size of mask in pixels */
		short linc; 			/* line increment */
		unsigned char base[2];	/* start of actual data */
} Mask;


typedef struct {
		pt xs, yr,				/* tranformation matrix */
		   xr, ys,				/*	 /	xs	yr	0 \  */
		   xd, yd;				/*	 |	xr	ys	0 |  */
} CTM;							/*	 \	xd	yd	1 /  */

typedef struct CTMel_ {
		struct CTMel_ *next;	/* pointer to next */
		CTM matrix; 			/* actual CTM */
} CTMel;

typedef struct _boxlist {
		struct _boxlist *next;	/* next box in list */
								/* previous values for */
		PPageBbox pagebbox; 	/* pagebbox, size of drawable area */
		PBbox clip; 			/* clipping box for bitmap */
		coor Origin;			/* page origin */
		coor ppp;				/* points per pixel (for conversion) */
		coor todev; 			/* pixels per point in default user space */
		pt Scale;				/* Scale cfr PFScalePage */
		CTM todefaultCTM;		/* CTM to convert to default user space */
} boxlist;

#include "font.h"
#include "CharPath.h"
#include "driver.h"

typedef struct _CachedFont CachedFont;
typedef struct _CachedMask CachedMask;

typedef struct _Bitmap {
		unsigned char *base;	/* base of memory with bitmap */
		short linc; 			/* line increment */
		int pinc;				/* plane increment */
		PBbox all;				/* bbox of max area */
		PBbox clip; 			/* bbox of area to draw in */

		PBbox patternbox;		/* size and offset (org) of current pattern */
		short pattlinc; 		/* line increment for pattern */
		int pattpinc;			/* plane increment for pattern */
		unsigned char *pattern; /* current pattern */

		short xcursor, ycursor; /* current cursor position */
		unsigned char *cursorpos;	/* cursor position in bitmap */
		short cursorbit;			/* offset in cursorpos where cursor is */
		unsigned int cursormask;	/* mask to set bit where cursor is */

		void *extra;			/* extra data for Bitmap driver */
		BitDrawRouts *bitdraw;	/* current drawing routs */
};

typedef struct _ColourCache {
		/* ColourCache entries are order in order of usage */
		/* (first is most recently used, last least recently used) */
		/* The list is circular and contains a (empty) header */
		struct _ColourCache *prev, *next;
		ColourCMYK colour;
		PBbox patternbox;		/* size and offset (org) of pattern */
		unsigned char pattern[2];
} ColourCache;

#define SaveableGstate \
		struct _gstatelist *next;	/* next entry in saved list */			\
		coor ppp;				/* points per pixel (for conversion) */ 	\
		coor todev; 			/* pixels per point in default user space */\
		PPageBbox pagebbox; 	/* device coordinates, max area in raster	\
								   which can be used - for this pass */ 	\
	/* added 04.11.2000, fontsize multiplier, needed for anti-aliased fonts */\
		int fontsizemultiplier; /* multiplier for fontsize, "set" by driver */\
	/* end of public part of Gstate */										\
		PBbox rasterclip;		/* copy of raster->clip during save */		\
	/* positioning etc. */													\
		coor Origin;			/* current page origin */					\
		pt Scale;				/* Scale cfr PFScalePage */ 				\
		CTM todefaultCTM;		/* CTM to convert to default user space */	\
	/* info for rendering paths */											\
		pt flatness;			/* accuracy for flattening curves */		\
		coor posdev;			/* current position in device coordinates */\
		coor subpathorg;		/* origin of subpath, used for closepath */ \
		pt linewidth;			/* width of a line (user space !!) */		\
		char pathtype;			/* method to draw path */					\
		char hairline;			/* TRUE when strokes should be hairline */	\
		segment *clippath;		/* current clipping path (linked by clip) */\
								/* always uses winding rule */				\
		char clippathmine;		/* TRUE -> release clippath on Restore */	\
	/* font info */ 														\
		char currfontname[PF_MAXFONTNAME];	/* name of current font */		\
		pt fontsize;			/* current size of font */					\
		char usecache;			/* do we have to use the fontcache ?? */	\
	/* draw & paper colour */												\
		char drawcolourspace;	/* colourspace for drawing colour */		\
		char papercolourspace;	/* colourspace for paper colour */			\
		ColourCMYK DrawColour;	/* drawing colour */						\
		ColourCMYK PaperColour; /* paper colour */							\
	/* parameters for drawing stroked lines */								\
		char linexmid,lineymid; /* TRUE when coors round to centers */		\
		pt linexppp, lineyppp;	/* conversion to calculate line length */	\
		char uniformgrid;		/* TRUE when x/y coordinates can be mixed */\
		pt linex2y, liney2x;	/* conversion between x and y coordinates */\
	/* statusses for handling strings */									\
		char printescape;		/* print or handle escape sequence ?? */	\
		char extraeos;			/* extra end-of-string character */ 		\
	/* pointer to colour table for PFPictureDisplayXXX */					\
		union {                                                             \
			pt *gray;														\
			ColourRGB *RGB; 												\
			ColourCMYK *CMYK;												\
		} pictcoltable; 													\
		int pictcolcurr;		/* currently used colour */ 				\
	/* pattern for colour distribution when building fastpatterns */		\
		char patternmask[8][8]

typedef struct _gstatelist {
		SaveableGstate;
} gstatelist;

struct _Gstate {
	/* NOTE all coordinates in a Gstate are in device space (absolute) */

	/* start of Gstate definition is public - any change here should also */
	/* be made in the PFmodule.h header file !! */

		char flag[15];			/* should be 'PROforma Gstate' */

		BITMdriver *bitmap; 	/* bitmap driver for this Gstate */
		DRIVdriver *printer;	/* printer driver for this Gstate */

		struct _DriverInfo *PrinterExtra;	/* extra data for Printer driver */
		Bitmap raster;			/* all information of the area to draw in */

	/* start of area saved/restored with PFGstateSave and PFGstateRestore */
	/* first part is still in public part of the gstate */
		SaveableGstate;
	/* end of PFGstateSave/Restore area */

		char BitmapAllocated;	/* TRUE when gstate allocated the bitmap */
		void *rasterbase;		/* rasterbase as allocated */
		void *rasterpattern;	/* fastpattern as allocated */

		PPageBbox allbbox;		/* pagebbox of entire page */
		PPageBbox firstbbox;	/* pagebbox for first band to render */

		int passes; 			/* number of passes to render 1 copy of page */

		DRIVdriver *nextdriver; /* for PFDriverNext command */
		int nextdrivernumber;	/* for PFDriverNext command */
		PICTdriver *nextpicture;/* for PFPictureNext command */
		int nextpicturenumber;	/* for PFPictureNext command */
		BITMdriver *nextbitmap; /* for PFBitmapNext command */
		int nextbitmapnumber;	/* for PFBitmapNext command */
		ColourCache *ColCache;	/* colour cache */
	/* transformation matrices, and other data to flatten paths */
		coor orgppp;			/* ppp	 when Scale==pt_one */
		coor orgtodev;			/* todev when Scale==pt_one */
		
		CTM currCTM;			/* current CTM */
		CTMel *CTMlist; 		/* list of CTM's for save and restore */
		
	/* path information, always flattened */
		segment *path;			/* current path */
								/* sorted, double linked, header & sentinel */
								/* horizontal lines are not included */
		
		segment *last;			/* last segment linked in (used to link in) */
		Chunks freesegments;	/* for allocating segments */
		segment *path2, *last2; /* for second pass in character rendering */
	/* font information */
		char noDisplayChars;	/* don't draw chars when TRUE */
		fontrec *fontlist;		/* list of all loaded fonts for this Gstate */
		font *currfont; 		/* current font */
		coor CharWidth; 		/* advance width for last displayed character */
		pt chwx;				/* width of char */
		CachedFont *font;		/* the current CachedFont - can be NULL */
		fontline *nextfont; 	/* counter for PFNextFontName */
		int boxxmin,boxymin;	/* minima for character bbox */
		int boxxmax,boxymax;	/* maxima for character bbox */
		CharPath chp;			/* for PFInitCharPath & CharPathEl */
		MutexHandler *CacheForce;	/* used to release cache on force remove */
	/* info for PFGstateSave and PFGstateRestore */
		gstatelist *savelist;	/* list of saved Gstates */
		gstatelist *freesavelist;	/* empty nodes for saving gstates */
	/* SubWindow info */
		boxlist *subwindows;	/* list of previous windows */
		boxlist *freesubwins;	/* list of old subwindows to prevent alloc */
};
